% demo_denoising - image denoising experiments in which the image is reordered using patch-based
% permutations and then filtered using learned filters.
% demo_denoising generates the image denoising results reported in Table III in:
% I. Ram, M. Elad, and I. Cohen, "Image Denoising Using NL-Means Via Smooth Patch Ordering", 
% Proc. 38th IEEE Internat. Conf. Acoust. Speech Signal Process., ICASSP-2013, Vancouver, Canada, May 26-31, 2013, pp. 1350-1354.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% September 2014
% =====================================================================================

clear
close all
clc

%%%% choosing the test image %%%%

picNumber = 1; % choose what image to denoise out of 3 images
if picNumber == 1
    picName = 'house.png';
elseif picNumber == 2 
    picName = 'Lena512.png';
elseif picNumber == 3
    picName = 'barbara.png';
end

%%%% parameters %%%%
%%%% N and B - number of rows in a square patch and square search area, used in the first iteration.
%%%% N2 and B2 - number of rows in a square patch and square search area, used in the second iteration.
%%%% C, eps1, Q_s, Q_e, gamma_s, gamma_e - first iteration design paramters
%%%% C2, eps2, Q_s2, Q_e2, gamma_s2, gamma_e2 - second iteration design paramters

sigma_n = 10; % choose sigmn_n value out of [ 10 , 25 , 50 ]
permNum = 10; % number of calculated permutations
if sigma_n == 10
    N = 7;  B = 31;   C = 1.2;  eps1 = 10;   Q_s = 9;   Q_e = 5;   gamma_s = 3.3;  gamma_e = 1.7; % first iteration parameters
    N2 = 4; B2 = 231; C2 = 1.1; eps2 = 10^3; Q_s2 = 33; Q_e2 = 3;  gamma_s2 = 0.4; gamma_e2 = 1.4; % second iteration parameters
elseif sigma_n == 25
    N = 12; B = 31;   C = 1.1;  eps1 = 10^2; Q_s = 11;  Q_e = 5;   gamma_s = 4.1;  gamma_e = 1.7; % first iteration parameters
    N2 = 4; B2 = 131; C2 = 0.3; eps2 = 10^7; Q_s2 = 71; Q_e2 = 11; gamma_s2 = 0.3; gamma_e2 = 0.5; % second iteration parameters
elseif sigma_n == 50
    N = 16; B = 31;   C = 1.1;  eps1 = 10^2; Q_s = 11;  Q_e = 5;   gamma_s = 5;    gamma_e = 5.5; % first iteration parameters
    N2 = 6; B2 = 141; C2 = 0.1; eps2 = 10^3; Q_s2 = 91; Q_e2 = 19; gamma_s2 = 0.2; gamma_e2 = 0.3; % second iteration parameters
end

%%%% reading test images %%%%

origPic = double( imread( [ 'images/' , picName ] ) );
[ N_m , N_n ] = size( origPic );

%%%% creating noisy images %%%%

rand('seed', 0);
randn('seed', 0);
randNoiseMat = sigma_n * randn( N_m , N_n );
noisyPic = origPic + randNoiseMat;
mseIn = mse( noisyPic - origPic );
PSNRIn = 20*log10( 255 ./ sqrt( mseIn ) );

%%%% Iteration 1 %%%%

randVec = rand( permNum , 2 );
randVec2 = rand( N_m * N_n , permNum , 2 );
recPic = denoiseNLMeans1D( noisyPic , sigma_n , permNum , N , C , B , eps1 , randVec , randVec2 , Q_s , Q_e , gamma_s , gamma_e );
mseOut = mse( recPic - origPic );
PSNROut = 20 * log10( 255 ./ sqrt( mseOut ) );

figure
subplot( 131 ); imshow( noisyPic , [ 0 , 255 ] , 'border' , 'tight' );
subplot( 132 ); imshow( recPic , [ 0 , 255 ] , 'border' , 'tight' ); title( '1 iteration result' ); xlabel( [ 'PSNR=' , num2str( PSNROut ) ] )
drawnow

%%%% Iteration 2 %%%%

randVec = rand( permNum , 2 );
randVec2 = rand( N_m * N_n , permNum , 2 );
recPic2 = denoiseNLMeans1D( noisyPic , sigma_n , permNum , N2 , C2 , B2 , eps2 , randVec , randVec2 , Q_s2 , Q_e2 , gamma_s2 , gamma_e2 , recPic );
mseOut2 = mse( recPic2 - origPic );
PSNROut2 = 20*log10( 255 ./ sqrt( mseOut2 ) );

subplot( 133 ); imshow( recPic2 , [ 0 , 255 ] , 'border' , 'tight' ); title( '2nd iteration result' ); xlabel( [ 'PSNR=' , num2str( PSNROut2 ) ] )
drawnow
