% recPic =  denoiseNLMeans1D( noisyPic , sigma_n , permNum , N , C , B , eps1 , randVec , randVec2 , Q_s , Q_e , gamma_s , gamma_e , recPicOld  ) 
% Apply denoising using NL-Means via smooth patch ordering to the noisy image noisyPic
% =====================================================================================
% Input:
% noisyPic, origPic - the noisy and clean input images, respectively.
% N, B - number of rows in a square patch and square search area.
% permNum - number of used permutations.
% randVec , randVec2 - random variables used to calculate the permutations.
% eps1 , C , Q_s , Q_e , gamma_s , gamma_e - design paramters.
% sigma_n - noise standard deviation.
% recPicOld - initialization image .
% =====================================================================================
% Output:
% recPic - contains the denoised image recPic.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% September 2014
% =====================================================================================

function recPic =  denoiseNLMeans1D( noisyPic , sigma_n , permNum , N , C , B , eps1 , randVec , randVec2 , Q_s , Q_e , gamma_s , gamma_e , recPicOld )

[ N_m , N_n ] = size( noisyPic );

% extract patches

FMat = im2col( noisyPic , [ N , N ] , 'sliding' );
Weight = reshape( mxSubim2im( ones( N^2 * ( N_m - N + 1 )*( N_n - N + 1 ) , 1 ) , N , N_m ) , N_m , N_n );
if nargin > 13
    FMatRec = im2col( recPicOld , [ N , N ] , 'sliding' );
    X_l = FMatRec;
else
    X_l = FMat;
end
ixPic = reshape( 1 : ( N_m - N + 1 ) * ( N_n - N + 1 ) , [ N_m , N_n ] - N + 1 );
tempIx = ixPic(:);
[ cur_rows , cur_cols ] = ind2sub( [ N_m , N_n ] - N + 1 , tempIx );
ixMat = [ cur_rows , cur_cols ];
X_lLength = size( X_l , 2 );
x_restMat = [ X_l ; ones( 1 , X_lLength ) ; -1/2 * sum( X_l.^2 ) ];
x_1Mat = [ X_l ; -1/2 * sum( X_l.^2 ) ; ones( 1 , X_lLength ) ];

% classify patches into smooth patches and patches with texture or edges

tempIx = std( X_l )' < C * sigma_n;
X_lLength1 = sum( tempIx );
X_lLength2 = sum( ~tempIx );
clusterIx1 = find( tempIx == 1 );
clusterIx2 = find( tempIx == 0 );

sortedIxMat = zeros( X_lLength1 , permNum );
sortedIxMat2 = zeros( X_lLength2 , permNum );
for k = 1 : permNum
    
    disp( [ 'calculating permutation ' , num2str( k ) ' / ' , num2str( permNum ) ] )
    
    % find a permutation from smooth patches
    
    randStartIx1 = ceil( X_lLength1 * randVec( k , 1 ) );
    tempIx1 = 1 : size( X_l , 2 );
    tempIx1 = tempIx1( tempIx )';
    randStartIx = tempIx1( randStartIx1 );
    freePoints = int32( ~tempIx );
    sortedIxMat( : , k ) = ...
        double( mxCalcPermRand2Sets( x_1Mat( : ) , x_restMat( : ) , ixMat( : ) , freePoints , N^2 + 2 , X_lLength , X_lLength1 , N_m - N + 1 , randStartIx , B , randVec2( 1:X_lLength1 , k , 1 ) , eps1 ) );
    
    % find a permutation from patches with texture or edges
    
    randStartIx2 = ceil( X_lLength2 * randVec( k , 2 ) );
    tempIx2 = 1 : size( X_l , 2 );
    tempIx2 = tempIx2( ~tempIx )';
    randStartIx = tempIx2( randStartIx2 );
    freePoints = int32( tempIx );
    sortedIxMat2( : , k ) = ...
        double( mxCalcPermRand2Sets( x_1Mat( : ) , x_restMat( : ) , ixMat( : ) , freePoints , N^2 + 2 , X_lLength , X_lLength2 , N_m - N + 1 , randStartIx , B , randVec2( 1:X_lLength2 , k , 2 ) , eps1 ) );
end
[ ~ , findIx ] = sort( sortedIxMat );
[ ~ , findIx2 ] = sort( sortedIxMat2 );

% denoise smooth patches

disp( 'denoising smooth patches' )

blocks2 = zeros( X_lLength1 , N^2 );
Q_s_h = floor( Q_s /2 );
for t = 1 : X_lLength1
    tempIx3 = [];
    for tt = 1 : permNum
        tempIx3 = [ tempIx3 ; sortedIxMat( max( 1 , findIx( t ,  tt ) - Q_s_h ) : min( X_lLength1 , findIx( t ,  tt ) + Q_s_h ) , tt ) ];
    end
    tempIx3 = unique( tempIx3 );
    W = exp( 2 * x_1Mat( : , clusterIx1( t ) )' * x_restMat( : , tempIx3 ) / ( N^2 * sigma_n^2 * gamma_s^2 ) );
    blocks2( t , :  ) = FMat( : , tempIx3 ) * ( W / sum( W ) )';
end

% denoise patches with texture or edges

disp( 'denoising patches with texture or edges' )

blocks3 = zeros( X_lLength2 , N^2 );
Q_e_h = floor( Q_e /2 );
for t = 1 : X_lLength2
    tempIx3 = [];
    for tt = 1 : permNum
        tempIx3 = [ tempIx3 ; sortedIxMat2( max( 1 , findIx2( t ,  tt ) - Q_e_h ) : min( X_lLength2 , findIx2( t ,  tt ) + Q_e_h ) , tt ) ];
    end
    tempIx3 = unique( tempIx3 );
    W = exp( 2 * x_1Mat( : , clusterIx2( t ) )' * x_restMat( : , tempIx3 ) / ( N^2 * sigma_n^2 * gamma_e^2 ) );
    blocks3( t , : ) = FMat( : , tempIx3 ) * ( W / sum( W ) )';
end

% recontsruct image from patches

blocks = zeros( X_lLength , N^2 );
blocks( clusterIx1 , : ) = blocks2;
blocks( clusterIx2 , : ) = blocks3;
recPic = reshape( mxSubim2im( blocks( : ) , N , N_m ) , N_m , N_n ) ./ Weight;